r"""
The class XEtaProduct stores sumes of extended eta products.  Extended eta
products are products of powers of `\eta((M z + h) / N)` and `\Lambda_{M,N,h}(z)`.
Here `M`, `N`, and `h` are integers and `\Lambda_{M,N,h}` is the logrithmic
derivative of `\eta((M z + h) / N) / eta(z)`.
"""

from copy import copy

from sage.categories.commutative_algebras import CommutativeAlgebras
from sage.categories.commutative_rings import CommutativeRings
from sage.categories.monoids import Monoids
from sage.categories.pushout import ConstructionFunctor
from sage.combinat.dict_addition import dict_addition
from sage.functions.all import floor, sgn, exp
from sage.libs.pari.all import pari
from sage.matrix.all import matrix, identity_matrix
from sage.misc.all import prod
from sage.modular.all import ModularForms
from sage.modules.all import vector
from sage.monoids.monoid import Monoid_class
from sage.rings.all import PolynomialRing, PowerSeriesRing, QQ, ZZ, Integer, infinity, CyclotomicField, CLF, ComplexField
from sage.rings.all import lcm, gcd, jacobi_symbol
import sage.rings.complex_field
from sage.rings.ring import CommutativeAlgebra
from sage.sets.all import PositiveIntegers, Family, Set
from sage.structure.element import MonoidElement, AlgebraElement
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.symbolic.all import SR, pi, I

import operator


################################################################################
### Shortcuts for extended eta products
################################################################################

def eta_lambda(l) :
    return XEtaProductAlgebra(l, QQ)( XEtaProducts()((0, l, 1, 0)) )

def eta_shift(l) :
    return XEtaProductAlgebra(l, QQ)( XEtaProducts()((1, l, 1, 0)) )

def etainv_shift(l) :
    return XEtaProductAlgebra(l, QQ)( XEtaProducts()((-1, l, 1, 0)) )

def eta_product(powers, shifts) :
    if not isinstance(shifts, (list, tuple)) :
        shifts = Integer(shifts).divisors()

    return prod( (eta_shift(s)**p if p > 0 else etainv_shift(s)**(-p)
                  for (s,p) in zip(shifts, powers) if p != 0),
                 XEtaProductAlgebra(1).one() )

def delta() :
    return eta_shift(1)**24

################################################################################
### Extended eta products
################################################################################

class XEtaProduct_class(MonoidElement) :
    r"""
    An extended eta product. This is a sum of products of powers of 
    `\eta((M z + h) / N)` and `\Lambda_{M,N,h}(z)`.  Here `M`, `N`,
    and `h` are integers and `\Lambda_{M,N,h}` is the logrithmic
    derivative of `\eta((M z + h) / N) / eta(z)`.


    ALGORITHM:

    The class stores its value in self._dict. Each of its items represents
    an extended eta product. The values are elements in a common (commutative)
    ring, that admits multiplication with roots of unity (as elements of a
    cyclotomic field).
     
    The keys are tuples of `4`-tuples.  Each of these tuples has the form
    ``(log, M, N, h)``.

    :

    - ``log``  -- Is either `0`, `-1`, or `1`.  Here, `0` stands for `\Lambda_{M,N,h}(z)`,
                  `-1` stands for `\eta((M z + h) / N)^{-1}`, and `1` stands for
                  `\eta((M z + h) / N)`.

    - `M`, `N` -- Positive integers.

    - `h`      -- An integer.

    EXAMPLE::

        sage: 
    """
    def __init__(self, e, parent = None, clean = True) :
        r"""
        INPUT:

        - `e` -- A tuple of tuples ``(log, M, N, h)``.  See the class' description.

        - ``parent`` -- An eta product algebra.

        - ``clean`` -- If ``True``, then sort `e` and remove unneseccary elements.
        """
        if parent is None :
            parent = XEtaProducts()

        MonoidElement.__init__(self, parent)

        self._e = e
        if clean :
            self._e = sorted(self._e)
            ne = list(self._e)

            for (log, M, N, h) in self._e :
                if log != -1 :
                    break

                try :
                    ne.remove((-log, M, N, h))
                    ne.remove((log, M, N, h))
                except :
                    pass
            self._e = tuple(ne)

    def _maximal_q_denominator(self) :
        r"""
        The maximal denominator of the exponent of `q` in the Fourier expansion
        of the extended eta product.

        OUTPUT:

        - An integer.

        EXAMPLES::
        
            sage: eps = XEtaProducts()
            sage: eps((0,1,3,0))._maximal_q_denominator()
            3
            sage: eps((1,1,1,0))._maximal_q_denominator()
            24
        """
        return lcm( sum((log * Integer(M) / (24 * N) for (log, M, N, _) in self._e), Integer(0)).denominator(),
                    lcm([N for (_, _, N, _) in self._e]) )

    def _zeta_denominator(self) :
        r"""
        The maximal exponent of root of unity that ocurr in the Fourier expansion of ``self``.

        EXAMPLES::
        
            sage: XEtaProducts()((0,1,3,1))._zeta_denominator()
            3
        """
        return _zeta_denominator_xep(self._e)

    def _square_normalization(self) :
        r"""
        The normalizing factor of an extended eta product.
        See ``_square_normalization_xep`` for more details.
        """
        return _square_normalization_xep(self._e)

    def weight(self) :
        r"""
        The weight of the extended eta proudct, if it is well-defined.

        OUTPUT:

        - An integer.

        EXAMPLES::
        
            sage: eps = XEtaProducts()
            sage: eps((0,3,1,0)).weight()
            2
        """
        if len(self._e) == 0 :
            return 0

        return    2 * len(filter(lambda (log, M, N, h): log == 0, self._e)) \
                + Integer(1)/2 * len(filter(lambda (log, M, N, h): log == 1, self._e)) \
                - Integer(1)/2 * len(filter(lambda (log, M, N, h): log == -1, self._e))
        

    def rescale(self, scale) :
        r"""
        UNTESTED!!
        Rescale the extended eta products `f(\tau)`, returning
        `f(scale * \tau)`.

        INPUT:

        - ``scale`` - A positive integer.
        """
        return self.parent().element_class( tuple( (log, scale * M, N, h)
                                                   for (log, M, N, h) in self._e ),
                                            self.parent() )

    def eval_projection(self, prec, cyclotomic_field) :
        r"""
        INPUT:
        
        See `:meth:eval_projection__by_modular_forms`.

        OUTPUT:

        - A power series in `q`.

        EXAMPLES::
        
            sage: eps = XEtaProducts()
            sage: eps((0,3,1,0)).eval_projection(10, QQ)
            1/4 + 3*q + 9*q^2 + 3*q^3 + 21*q^4 + 18*q^5 + 9*q^6 + 24*q^7 + 45*q^8 + 3*q^9 + O(q^10)
        """
        if prec < 0 :
            raise ValueError( "Precision must be non-negative." )

        return _eval_projection_xep(self._e, prec, cyclotomic_field)

    def _mul_(self, other) :
        r"""
        """
        return self.parent().element_class(tuple(sorted(self._e + other._e)), self.parent())

    def __invert__(self) :
        r"""
        """
        if any(log == 0 for (log, _, _, _) in self._e) :
            raise ValueError( "Inverse of {0} is not an extended eta product.".format(self) )

        return self.parent().element_class( tuple(sorted((-log, M, N, h) for (log, M, N, h) in self._e)), self.parent() )
        
    def __cmp__(self, other) :
        r"""
        """
        c = cmp(type(self), type(other))

        if c == 0 :
            c = cmp(self._e, other._e)

        return c

    def _repr_(self) :
        r"""
        EXAMPLES::
        
            sage: eps = XEtaProducts()
            sage: eps((0,3,3,1))**3
            Lambda_{3,3,1}(z)^3
        """
        if len(self._e) == 0 :
            return "eta(z)^0"

        terms = list()
        for (log, M, N, h) in sorted(Set(self._e), key = lambda (log, M, N, e): (abs(log), M, N, e)) :
            if log == 0 :
                term = "Lambda_{{{0},{1},{2}}}(z)".format(M, N, h)
            else :
                if h != 0 and N != 1 :
                    term = "eta(({0}z + {2})/{1})".format("" if M ==  1 else M, N, h)
                elif h != 0 :
                    term = "eta({0}z + {2})".format("" if M ==  1 else M, N, h)
                elif N != 1 :
                    term = "eta({0}z/{1})".format("" if M ==  1 else M, N)
                else :
                    term = "eta({0}z)".format("" if M ==  1 else M)

            c = self._e.count((log, M, N, h))
            if log == -1 :
                term += "^-{0}".format(c)
            elif c != 1 :
                term += "^{0}".format(c)
            
            terms.append(term)
            
        return ' * '.join(terms)

    def __hash__(self) :
        r"""
        """
        return hash(self._e) + 117

class XEtaProducts(UniqueRepresentation, Parent) :
    r"""
    The set of extended eta products.  An extended eta product is
    a product of powers of `\eta((M z + h) / N)` and `\Lambda_{M,N,h}(z)`.
    Here `M`, `N`, and `h` are integers and `\Lambda_{M,N,h}` is the logrithmic
    derivative of `\eta((M z + h) / N) / eta(z)`.

    EXAMPLES::
    
        sage: eps = XEtaProducts()
        sage: eps((1,3,1,0))**-2 * eps((1,9,1,0))**6
        eta(3z)^-2 * eta(9z)^6
        sage: eps((-1,3,1,0))**2 * eps((1,9,1,0))**6
        eta(3z)^-2 * eta(9z)^6
    """

    Element = XEtaProduct_class

    def __init__(self) :
        r"""
        """
        Parent.__init__(self, category = Monoids())

    def _an_element_(self) :
        r"""
        """
        return self( ((1,2,1,0), (0,7,3,2)) )

    def _element_constructor_(self, e) :
        r"""
        INPUT:
        
        - `e` -- Either `1`, a tuple ``(log, M, N, h)``, or a list or tuple of such
                 tuples.  See `:class:XEtaProduct_class` for a description of
                 ``log``, `M`, `N`, and `h`.

        EXAMPLES:
        
            sage: eps = XEtaProducts()        
            sage: eps(1)
            eta(z)^0
            sage: eps((1,9,1,0))
            eta(9z)
            sage: eps([(1,9,1,0)]) == eps(((1,9,1,0),))
            True
        """
        if e == 1 :
            return self.element_class(tuple(), self)

        if isinstance(e, tuple) and not isinstance(e[0], tuple) :
            e = [e,]

        if not isinstance(e, (tuple, list)) or not all( log in [-1,0,1] and M in ZZ and M > 0
                                                 and N in ZZ and N > 0 and h in ZZ
                                                for (log, M, N, h) in e ) :
            raise TypeError( "e (={0}) does not represent an extended eta product".format(e) )
        
        return self.element_class(tuple(e), self)

    def _repr_(self) :
        r"""
        EXAMPLES::

            sage: XEtaProducts()
            Set of all extended eta products
        """
        return "Set of all extended eta products"

################################################################################
### XEtaProductAlgebra
################################################################################

class XEtaProductSum(AlgebraElement) :
    def __init__(self, d, parent, clean = True, check = False) :
        r"""
        INPUT:
        
        - `d`        -- A dictionary.  The coefficients must lie in the base ring
                        of ``parent``.

        - ``parent`` -- An `:class:XEtaProductAlgebra`.
        """
        AlgebraElement.__init__(self, parent)

        self._dict = copy(d)

        if clean :
            for (e, v) in list(self._dict.iteritems()) :
                if v == 0 :
                    del self._dict[e]
        if check :
            for v in self._dict.values() :
                if v.parent() is not self.parent().base_ring() :
                    raise ValueError( "Coefficients must be elements of {0} (but the parent of {1} is {2})" \
                                          .format(self.parent().base_ring(), v, v.parent()) )

    def _maximal_q_denominator(self) :
        r"""
        The maximal denominator of the exponent of `q` in the Fourier expansion
        of ``self``.

        OUTPUT:

        - An integer.
        """
        return lcm([ e._maximal_q_denominator() for e in self._dict.keys() ])

    def _zeta_denominator(self) :
        r"""
        The maximal exponent of root of unity that ocurr in the Fourier expansion of ``self``.
        """
        eta_zeta_denominator = lcm([ e._zeta_denominator() for e in self._dict.keys() ])
        coefficient_zeta_denominator = lcm([c.parent().zeta_order() for c in self._dict.values()])

        return lcm(eta_zeta_denominator, coefficient_zeta_denominator)

    def weight(self) :
        r"""
        The weight of the extended eta proudct, if it is well-defined.  In all
        other cases ``None``.

        OUTPUT:

        - An integer or ``None``.
        """
        if len(self._dict) == 0 :
            return 0

        weights = [e.weight() for e in self._dict.keys()]
        if len(set(weights)) == 1 :
            return weights[0]
        else :
            return None

    def rescale(self, scale) :
        r"""
        UNTESTED!!
        Rescale the extended eta products `f(\tau)`, returning
        `f(scale * \tau)`.

        INPUT:

        - ``scale`` - A positive integer.
        """
        parent = XEtaProductAlgebra(self.parent().level() * scale)

        return parent.element_class( dict( (e.rescale(scale), v) for (e, v) in self._dict.iteritems() ),
                                                     parent )

    def _eval_projection__check_cache(self, prec) :
        r"""
        Check the cache for an evaluation of the extended eta product.

        INPUT:
        
        - ``prec -- An integer.
        """
        if prec < 0 :
            raise ValueError( "Precision must be non-negative." )

        if not hasattr(self, "_eval_projection__cache") :
            self._eval_projection__cache = (0, PowerSeriesRing(QQ, 'q')(0))

        if prec <= self._eval_projection__cache[0] :
            return self._eval_projection__cache[1].truncate_powerseries(prec)

        if len(self._dict) == 0 :
            self._eval_projection__cache = (infinity, PowerSeriesRing(QQ, 'q')(0))

            return self._eval_projection__cache[1].truncate_powerseries(prec)
        
        raise ValueError( "Evaluation not cached." )

    def _eval_projection__primitive(self, prec, cyclotomic_field) :
        r"""
        INPUT:
        
        See `:meth:eval_projection`.

        OUTPUT:

        - A power series in `q`.  The base ring is not necessarily ``cyclotomic_field``.
        """
        try :
            return self._eval_projection__check_cache(prec)
        except ValueError :
            pass

        if cyclotomic_field is None :
            zeta_denominator = self._zeta_denominator()

            if zeta_denominator in [1, 2] :
                cyclotomic_field = QQ
            else :
                cyclotomic_field = CyclotomicField(zeta_denominator)

        result = sum(v * e.eval_projection(prec, cyclotomic_field) for (e, v) in self._dict.iteritems())

        if cyclotomic_field is None or not isinstance(cyclotomic_field, sage.rings.complex_field.ComplexField_class) :
            self._eval_projection__cache = (prec, copy(result))

        return result
        
    def eval_projection(self, prec, cyclotomic_field = None) :
        r"""
        Eval `q`-expansion of the extended eta product and project it to the
        part which has integral `q`-exponents.
        
        NOTE:
        
        The algorithm assumes that the extended eta product is a modular form of
        given level.

        INPUT:
        
        - ``prec`` -- An integer. The precision of the result.
        
        - ``cyclotomic_field`` -- A cyclotomic field, a complex numerical field, or ``None``.
                                  In the first case the field must contain all roots of unity
                                  that occur during the calculation.

        OUTPUT:

        - A power series in `q`.

        ALGORITHM:
        
        Eval the projection up to sufficient precision, so that the modular form is determined 
        unquely.  Then compute the rest of the expansion using the stand implementation
        of modular forms.
        """
        try :
            return self._eval_projection__check_cache(prec)
        except ValueError :
            pass

        ## TODO: Get graded parts and then try again
        if self.weight() is None :
            raise NotImplementedError( "Can only evaluate elements of homoegeneous weight." )

        if self.weight() == 0 :
            if len(self._dict) == 1 and self._dict.keys()[0] == XEtaProducts()(1) :
                self._eval_projection__cache = \
                  (infinity, PowerSeriesRing(self._dict.values()[0].parent(), 'q')(self._dict.values()[0]))

                return self._eval_projection__cache[1].truncate_powerseries(prec)
                

        if self.parent().level() is None :
            return self._eval_projection__primitive(prec, cyclotomic_field)


        mf = ModularForms(self.parent().level(), self.weight())
        if mf.dimension() == 0 :
            self._eval_projection__cache = (infinity, PowerSeriesRing(QQ, 'q')(0))

            return self._eval_projection__cache[1].truncate_powerseries(prec)

        pre_prec = max([ b.valuation() for b in mf.echelon_basis() ]) + 1
        pre_expansion = self._eval_projection__primitive(pre_prec, cyclotomic_field)

        result = sum( ( pre_expansion[b.valuation()] * b.qexp(prec)
                        for b in mf.echelon_basis() ),
                      pre_expansion.parent().zero_element() )

        if cyclotomic_field is None or not isinstance(cyclotomic_field, sage.rings.complex_field.ComplexField_class) :
            self._eval_projection__cache = \
                ( prec, copy(result) )

        return result

    def cusp_expansion(self, c) :
        r"""
        Return the representation of the extended eta product at a cusp `c`.

        NOTE:
        
        The expansion will lie in the same parent as ``self``.

        INPUT:

        - `c` -- A cusp.
        """
        if self == 0 :
            return self.parent().zero()

        cmat = _cusp_matrix(c)

        ndict_r = list()
        for (e, v) in self._dict.iteritems() :
            if e == XEtaProducts().one() :
                ndict_r.append((e, v, QQ(0)))
            else :
                (ne, roots) = zip(*[ _cusp_expansion(eta, cmat) for eta in e._e ])
                ndict_r.append((XEtaProducts()(ne), v, sum(roots)))

        zeta_denominator = lcm([r.denominator() for (_,_,r) in ndict_r])
        ## we also need to use cyclotomic fields in order to express the normalization
        for (e, _, _) in ndict_r :
            n = QQ(e._square_normalization())
            n = n.squarefree_part()
            if n != 1 :
                zeta_denominator = lcm(zeta_denominator, 4 * n)

        if 2 % zeta_denominator == 0 :
            ndict = [(e, e._square_normalization().sqrt() * (1 if r in ZZ else -1) * v) for (e, v, r) in ndict_r]
            K = self.parent().base_ring()
        else :
            ## There is a bug in coercion of cyclotomic fields that we work around here.
            ## We assume that K is a cyclotomic field with standard embedding
            import sage.rings.number_field.number_field
            cyc_fields = filter( lambda p: not isinstance(p, sage.rings.rational_field.RationalField),
                                 [v.parent() for (_,v,_) in ndict_r] )
            if not all( isinstance(p, sage.rings.number_field.number_field.NumberField_cyclotomic)
                        for p in cyc_fields ) :
                raise NotImplementedError( "Workaround for coercion only available for cyclotomic fields with standart embedding." )

            if not all( p.coerce_embedding()(p.gen())
                        == CLF(exp(2 * pi * I / p.gen().multiplicative_order()))
                        for p in cyc_fields ) :
                raise NotImplementedError( "Workaround for coercion only available for cyclotomic fields with standart embedding." )

            v_order = lcm([1] + [ p.gen().multiplicative_order() for p in cyc_fields ])

            cyclotomic_order = lcm([ zeta_denominator, v_order,
                                     self.parent().base_ring().zeta_order() ])
            K = CyclotomicField(cyclotomic_order)
            zeta = K.gen()

            to_cyclotomic = lambda c: \
                K(QQ(c)) \
                if c in QQ else \
                sum( cc * zeta**p
                     for (cc, p) in zip(c.list(), range(0, cyclotomic_order,
                                                        cyclotomic_order // c.parent().gen().multiplicative_order()))
                     if cc != 0 )
            R = PolynomialRing(K, 'x')
            x = R.gen()

            ndict = list()
            for (e, v, r) in ndict_r :
                n = QQ(e._square_normalization())
                nsf = n.squarefree_part()
                n = n / nsf
                if nsf != 1 :
                    nsf = (x**2 - nsf).factor()[0][0][0]
                    if nsf.complex_embedding() < 0 :
                        nsf = -nsf

                ndict.append( (e, n.sqrt() * nsf * zeta**(r * cyclotomic_order) * to_cyclotomic(v)) )
                
        return self.parent().construction()[0](K)(dict( (e, v) for (e,v) in ndict ))

    def _add_(self, other) :
        r"""
        """
        return self.parent().element_class(dict_addition([self._dict, other._dict]), self.parent())

    def _mul_(self, other) :
        r"""
        """
        return self.parent().element_class( dict( (e * eo, v * vo)
                                                  for (e, v) in self._dict.iteritems()
                                                  for (eo, vo) in other._dict.iteritems()),
                                            self.parent() )

    def __cmp__(self, other) :
        r"""
        """
        return cmp(self._dict, other._dict)

    def _repr_(self) :
        r"""
        """
        return ' + '.join(repr(v) + "*" + repr(e) for (e, v) in self._dict.iteritems())

    def __hash__(self) :
        r"""
        """
        return hash(tuple(self._dict.keys())) + hash(tuple(self._dict.values()))

class XEtaProductAlgebra(UniqueRepresentation, CommutativeAlgebra) :
    r"""
    NOTE:
    
    We don't use CombinatorialAlgebra, which seems depreciated.
    """
    Element = XEtaProductSum

    def __init__(self, level = 1, base_ring = QQ) :
        r"""
        The elements of this algebra are assumed to be correspond to 
        modular forms of given level.

        INPUT:

        - ``level`` -- A positive integer or ``None``, which means all
                       levels.
        
        - ``base_ring`` -- A commutative ring, that is the coefficient ring
                           of the algebra.
        """
        CommutativeAlgebra.__init__(self, base_ring, category = CommutativeAlgebras(base_ring))

        self._level = level

        self._populate_coercion_lists_(coerce_list=[base_ring])

    def level(self) :
        r"""
        The level of extended eta products in this algebra.
        
        OUTPUT:
        
        - A postive integer.
        """
        return self._level

    def _an_element_(self) :
        r"""
        """
        if self.level() is None :
            return self.element_class( {XEtaProducts().an_element() : self.base_ring().an_element()}, self)
        else :
            return self.element_class( {XEtaProducts()((1,self.level(),1,0))**24 : self.base_ring().an_element()}, self)

    def construction(self) :
        return (XEtaProductAlgebra_functor(self._level), self.base_ring())

    def _coerce_map_from_(self, other) :
        if isinstance(other, XEtaProductAlgebra) and \
           self.base_ring().has_coerce_map_from(other.base_ring()) :
            return self._level is None or \
                (other._level is not None and
                 Integer(other._level).divides(self._level))
    
    def _element_constructor_(self, x) :
        r"""
        INPUT:
        
        - `x` -- We treat the following cases:
                 - An element of the base ring.
                 - A dictionary of type XEtaProducts -> coefficient.
        """
        if x in XEtaProducts() :
            return self.element_class( {x: self.base_ring().one()}, self )

        if isinstance(x, dict) and all(e in XEtaProducts() for e in x.keys()) :
            if not all(v in self.base_ring() for v in x.values()) :
                return self.element_class( dict( (e, self.base_ring(v)) for (e,v) in x.iteritems() ), self)
            return self.element_class( x, self)

        if x in self.base_ring() :
            return self.element_class( { XEtaProducts()(1) : self.base_ring()(x) }, self )
        
        if isinstance(x.parent(), XEtaProductAlgebra) and \
           self.base_ring().has_coerce_map_from(x.parent().base_ring()) :
            return self.element_class( dict((e, self.base_ring()(v)) for (e, v) in x._dict.iteritems()), self )

        raise TypeError( "Cannot convert x (={0}) into {1}.".format(x, self) )

    def _repr_(self) :
        return "Algebra of extended eta products{0} over {1}".format(" of level {0}".format(self._level), self.base_ring())

class XEtaProductAlgebra_functor( ConstructionFunctor ) :
    r"""
    A functor that create an algebra of eta products of given level.
    """

    rank = 10

    def __init__(self, level) :
        ConstructionFunctor.__init__(self, CommutativeRings(), CommutativeRings())
        
        self._level = level

    def merge(self, other) :
        if type(self) != type(other) :
            return None

        if self._level is None or other._level is None :
            return XEtaProductAlgebra_functor(None)

        return XEtaProductAlgebra_functor(lcm(self._level, other._level))

    def _apply_functor(self, R) :
        return XEtaProductAlgebra(self._level, R)


################################################################################
### Helper functions
################################################################################

### evaluation of eta products; helper functions

def _divide_q_power(s, d) :
    r"""
    From a power series `s = \sum a_n x^n` form another `\sum a_{d n} x^n`.

    INPUT:
    
    - `s`  -- A power series.
    
    - `d`  -- A positive integer.

    OUTPUT:
    
    -  A power series.
    """
    FRq = s.parent()
    if s.prec() is infinity :
        prec = s.degree() // d + 1
    else :
        prec = (s.prec() - 1) // d + 1
        
    res = FRq(dict((e // d, c) for (e, c) in s.dict().iteritems() if e % d == 0))
    if s.prec() is infinity :
        return res
    else :
        return res.add_bigoh(prec)

def _eta(prec) :
    r"""
    Return the `q`-expansion of `q^{-1/24} \eta(z)` with precision ``prec``.

    INPUT:

    - ``prec`` -- A positive integer.

    OUTPUT:

    - A power series in `q`.
    """
    pari.set_series_precision(prec)
    return PowerSeriesRing(QQ, 'q')(pari('eta(q)')).add_bigoh(prec)

def _eta_d(prec) :
    r"""
    Return the `q`-expansion of `(2 \pi i)^{-1} q^{-1/24} \partial_z \eta(z)` with precision ``prec``.

    INPUT:

    - ``prec`` -- A positive integer.

    OUTPUT:

    - A power series in `q`.
    """
    R = PowerSeriesRing(QQ, 'q')
    q = R.gen()
    return Integer(1) / 24 * _divide_q_power(R(_eta(prec)).subs(q = q**24).shift(1).derivative(), 24)

def _zeta_denominator_xep(e) :
    r"""
    The order of the root of unity that occurs in the evaluation of an extended eta product `e`.

    INPUT:
    
    - `e` -- A list of `4`-tuples that represents an extended eta product.

    OUTPUT:
    
    - An integer.
    """
    zeta_den = sum((log * Integer(h) / (24 * N) for (log, _, N, h) in e), QQ(0)).denominator()

    for (log, M, N, h) in e :
        if h == 0 :
            continue
        else :
            zeta_den = lcm(zeta_den, N)

    return zeta_den

### evaluation of eta products

def _eval_xep(e, prec, cyclotomic_field = None) :
    r"""
    Eval an extended eta product.  The result is to be viewed as a power series
    in `q^{1/lN}`.

    The cyclotomic field must contain all roots of unity that occur in the product
    expansion.

    INPUT:

    - `e` -- A list of `4`-tuples that represents an extended eta product.

    - ``prec`` -- An integer; The desired precision of the result (as a power
                  series in `q`).

    - ``cyclotomic_field`` -- Either ``None`` or a cyclotomic field.  In the
                              first case the result is a power series over `\bold{Q}`,
                              in the second case it is a power series over that field.

    OUTPUT:

    - A pair `(s, lN)`, where `s|_{q = q^{1/lN}` is the evaluation of `e`.
    """ 
    if cyclotomic_field is None :
        K = QQ
        zeta = -1
        zeta_ord = 2
    else :
        K = cyclotomic_field
        zeta = K.gen()
        zeta_ord = K.zeta_order()

        if zeta.multiplicative_order() != zeta_ord :
            zeta = -zeta
            assert zeta_ord % 2 == 0
            assert zeta**(zeta_ord // 2) != 1

    return _eval_xep__with_zeta(e, prec, K, zeta, zeta_ord)

def _eval_xep__complex(e, prec, cyclotomic_field) :
    r"""
    Eval an extended eta product using numerical multiplication.  The result
    is to be viewed as a power series in `q^{1/lN}`.

    The cyclotomic field must contain all roots of unity that occur in the product
    expansion.

    INPUT:

    - `e` -- A list of `4`-tuples that represents an extended eta product.

    - ``prec`` -- An integer; The desired precision of the result (as a power
                  series in `q`).

    - ``cyclotomic_field`` -- Either ``None`` or a cyclotomic field.  In the
                              first case the result is a power series over `\bold{Q}`,
                              in the second case it is a power series over that field.

    OUTPUT:

    - A pair `(s, lN)`, where `s|_{q = q^{1/lN}` is the evaluation of `e`.
    """ 
    c_prec = 10000
    coefficient_den = 24**len(filter(lambda (log, M, N, h): log == 0, e))
    CC = ComplexField(c_prec)


    if cyclotomic_field is QQ :
        cyc_basis_matrix = identity_matrix(CC, 1)
        
        evaluations = [ _eval_xep__with_zeta( e, prec, CC, CC(-1), 2 ) ]
    else :
        coprime_exps = Integer(cyclotomic_field.zeta_order()).coprime_integers(cyclotomic_field.zeta_order() + 1)

        cyc_basis_matrix = matrix(CC, [ map( operator.pow, len(coprime_exps) * [a.complex_embedding()], coprime_exps )
                                       for a in cyclotomic_field.power_basis() ]).transpose()
        evaluations = [ _eval_xep__with_zeta( e, prec, CC, CC(exp( 2 * SR('pi') * SR('I') * QQ((i, cyclotomic_field.zeta_order())) )), cyclotomic_field.zeta_order() )
                        for i in coprime_exps ]

        if any(ev[1] != evaluations[0][1] for ev in evaluations[1:]) :
            raise RuntimeError( "Scaling of numerically computed evaluations should coincide" )

    FEQ = PowerSeriesRing(cyclotomic_field, 'q')
    res = FEQ(0)
    for i in range(evaluations[0][0].prec()) :
        coeff = cyc_basis_matrix.solve_right(vector([coefficient_den * ev[0][i] for ev in evaluations]))
        res += FEQ.gen()**i * sum(map(operator.mul, [Integer(round(c.real_part())) for c in coeff], cyclotomic_field.power_basis())) / coefficient_den
    res = res.add_bigoh(evaluations[0][0].prec())

    return (res, evaluations[0][1])

def _eval_xep__with_zeta(e, prec, K, zeta, zeta_ord) :
    r"""
    Eval an extended eta product.  The result is to be viewed as a power series
    in `q^{1/lN}`.

    The field `K` must contain all roots of unity that occur in the product
    expansion.

    INPUT:

    - `e` -- A list of `4`-tuples that represents an extended eta product.

    - ``prec`` -- An integer; The desired precision of the result (as a power
                  series in `q`).

    - `K`    -- Either a cyclotomic field or an instance of ComplexField.

    - ``zeta`` -- An element of `K` that is a ``zeta_ord``-th root of unity.

    - ``zeta_ord`` -- A positive integer.

    OUTPUT:

    - A pair `(s, lN)`, where `s|_{q = q^{1/lN}` is the evaluation of `e`.
    """ 
    if len(e) == 0 :
        return PowerSeriesRing(K, 'q')(1).add_bigoh(bound)

    # The initial power of `q` attached the eta product. We will
    # add this later, but it must be non-negative.
    initial_power = sum(log * M / Integer(24 * N) for (log, M, N, _) in e)
    assert initial_power >= 0

    # The denominator of the powers of `q`, that we call `lN` in the documentation.
    lcm_N = lcm([N for (_, _, N, _) in e])

    FEQ = PowerSeriesRing(K, 'q')
    q = FEQ.gen()


    max_NbyM = max([Integer(N)/M for (_, M, N, _) in e])
    eta_max_prec = floor(max_NbyM * (prec - 1)) + 1
    if any(log == 0 for (log, _, _, _) in e) :
        eta_max_prec = max(eta_max_prec, lcm_N * (prec - 1) + 1)

    eta_b = _eta(eta_max_prec)

    if any(log == 0 or log == -1 for (log, _, _, _) in e) :
        eta_i = FEQ(eta_b**-1)

    if any(log == 0 for (log, _, _, _) in e) :
        eta_d = FEQ(_eta_d(eta_max_prec))
        eta_d_i = eta_d * eta_i

    eta_b = FEQ(eta_b)


    term = FEQ(1).add_bigoh(lcm_N * (prec - 1) + 1)
    for (log, M, N, h) in Set(e) :
        power = e.count((log, M, N, h))
        scale = M * lcm_N // N

        if zeta_ord * h % N != 0 :
            raise ValueError( "Roots of unit occur that do not lie in the coefficient field." )

        if log == 0:
            term *= ( M * N * (Integer(M) / N * eta_d_i(zeta**((zeta_ord * h) // N) * q**scale)
                                 - eta_d_i(q**lcm_N)) )**power
        elif log == 1 :
            term *= eta_b(zeta**((zeta_ord * h) // N) * q**scale)**power
        else :
            term *= eta_i(zeta**((zeta_ord * h) // N) * q**scale)**power

    if not initial_power * lcm_N in ZZ :
        lcm_N_n = lcm(initial_power.denominator(), lcm_N)
        term = term(q**(lcm_N_n // lcm_N))
        lcm_N = lcm_N_n
    term = term.shift(Integer(initial_power * lcm_N))


    # the additional root of unity that arises from applying q^{h/(24 * N)}
    zeta_exp = sum(log * Integer(h) / (24 * N) for (log, _, N, h) in e)
    if zeta_ord * zeta_exp not in ZZ :
        raise ValueError( "Roots of unit occur that do not lie in the coefficient field." )
    term *= zeta**(zeta_ord * zeta_exp)

    return (term.truncate_powerseries(lcm_N * (prec - 1) + 1), lcm_N)

def _eval_projection_xep(e, prec, cyclotomic_field = None) :
    r"""
    The procjection of an extended eta product to a modular form on `\Gamma_1(N)`,
    where `N` is the level of the extended eta product.

    The cyclotomic field must contain all roots of unity that occur in the product
    expansion.
    
    INPUT:

    See `func::_eval_xep`.

    OUTPUT:

    - A power series `s`.
    """
    ## To speed up computations use the following line.  The default precision used is 2000 bits, which
    ## should be sufficient for most applications.  But a detailed analysis of condition numbers etc. is up
    ## to you.
    ## (term, lcm_N) = _eval_xep__complex(e, prec, cyclotomic_field)
    (term, lcm_N) = _eval_xep(e, prec, cyclotomic_field)
    term = _divide_q_power(term, lcm_N)

    return term.truncate_powerseries(prec)

### Expansion of eta products at cusp

def _cusp_matrix(cusp) :
    r"""
    A matrix that transforms `\infty` to the given cusp.

    INPUT:

    - ``cusp`` -- Either a cusp or a rational.
    """
    s = cusp.numerator()
    u = cusp.denominator()
    (_, v, t) = Integer(s).xgcd(u)
    t = -t

    return (s, t, u, v)

def _cusp_expansion((log, M, N, h), (a, b, c, d)) :
    r"""
    This is taken from Koehler's book on eta products.
    
    INPUT:

    - ``(log, M, N, h)`` -- The data of an extended eta product; See
                            `class:XEtaProduct_class`. We assume that
                            `N = 1` and `h = 0`.

    - ``(a, b, c, d)``   -- The entries of a matrix in `\SL{2}(\ZZ)` which
                            the expansion is computed at.
    """
    if N != 1 or h != 0 :
        raise NotImplementedError( "Cusp expansions available only for N = 1 and h = 0." )

    ## to make things easier we use the notation by Koehler
    tmp = a
    a = d
    d = tmp
    b = -b
    c = -c

    g = gcd(M, c)
    x = (M * d) // gcd(M, c)
    u = - (c // gcd(M, c))
    (_, v, y) = Integer(x).xgcd(u)
    y = -y

    nu = - M * b * v - y * a

    t = gcd([g**2, M, nu * g])

    return ((log, g, M // g, nu), log * _eta_multiplier(x, y, u, v))

def _eta_multiplier(a, b, c, d) :
    r"""
    This is taken from Koehler's book on eta products.

    INPUT:

    - `a`, `b`, `c`, `d` -- The entries of a matrix in `\SL{2}(\ZZ)`.
    """
    if c % 2 == 0 :
        e = Integer(1)/24 * ((a + d) * c - b * d * (c**2 - 1) + 3 * d - 3 - 3 * c * d)
        if c == 0 :
            e += 0 if d == 1 else Integer(1)/2
        else :
            if jacobi_symbol(c, abs(d)) == -1 :
                e += Integer(1)/2
            e += Integer((sgn(c) - 1) * (sgn(d) - 1)) / 8
    else :
        e = Integer(1)/24 * ((a + d) * c - b * d * (c**2 - 1) - 3 * c)
        if d != 0 and jacobi_symbol(d, abs(c)) == -1 :
            e += Integer(1)/2

    return e

def _square_normalization_xep(e) :
    r"""
    The normalizing factor of an extended eta product.  This arises from the fact that
    `\eta(M/N \tau)` would normally carry a factor `(M/N)^{1/4}`.  That is, we apply elements
    of `\GL{2}(\ZZ)` in order to optain the eta products.

    We compute the square of this normalizing factor, which is always a rational.
    
    INPUT:

    - `e` -- A list of `4`-tuples that represents an extended eta product.

    OUTPUT:
    
    - An integer or a rational.
    """
    normalization = Integer(1)
    
    for (log, M, N, h) in e :
        if log == 1 :
            normalization /= N
        elif log == -1 :
            normalization *= N

    return normalization

