r"""
AUTHORS:

- Soumya Battachari, initial version of find_eta_products in GP

- Martin Raum, portage to Sage and other functions
"""
import cython
from sage.combinat.all import Subsets
from sage.modular.all import Cusp, FareySymbol, Gamma0
from sage.matrix.all import matrix, identity_matrix
from sage.functions.all import ceil
from sage.modules.all import span, vector
import operator

from eta_products import *
from eta_products import _eta_multiplier
from eta_products_helper_cython import _iterate_eta_products


def convert_to_eta_product_expression( f ) :
    r"""
    
    INPUT:

    - `f` -- A sum of extended eta products.

    OUTPUT:
    
    - A string.
    
    EXAMPLES:
    
        sage: convert_to_eta_product_expression( 2 * eta_product((2,5,-3), 4) - 1/4 * eta_product((2,2), 7) )
        '(   (Integer(-1)/4) * eta_product((2, 2) , 7)\n  + (2) * eta_product((2, 5, -3) , 4) )'
    """
    sum_strings = list()

    for (e, c) in f._dict.iteritems() :
        eta_level = 1
        product_strings = list()
        for (log, M, N, h) in e._e :
            if h != 0 :
                raise NotImplementedError()
            if N != 1 :
                raise NotImplementedError()

            if log == 0 :
                product_strings.append( "eta_lambda({0})".format(M) )
            else :
                eta_level = lcm(eta_level, M)

        if any( log != 0 for (log, _, _, _) in e._e ) :
            product_strings.append(
                "eta_product({exponents} , {eta_level})"
                .format( exponents = tuple(e._e.count((1, d, 1, 0)) - e._e.count((-1, d, 1, 0))
                                           for d in Integer(eta_level).divisors()),
                         eta_level = eta_level ) )

        coeff_sting = str(c) if c in ZZ else "Integer(" + str(c.numerator()) + ")/" + str(c.denominator())
        sum_strings.append( "(" + coeff_sting + ") * " + " * ".join(product_strings) )

    return "(   " + "\n  + ".join(sum_strings) + " )"

def span_with_eta_products(N, k, helper_level = 1, nmb_vertices = None, cache = None, loader = None) :
    r"""
    Span the space of level `N` and weight `k` modular forms by linear combinations
    of eta products which have level at most `N \cdot {\rm helper_level}`.  

    If the algorithm does not terminate, try setting ``nmb_vertices = 2000``.

    INPUT:
    
    - `N` -- A positive integer.

    - `k` -- A positive half-integer.

    - ``helper_level`` -- A positive integer.
    
    - ``nmb_vertices`` -- A positive integer or ``None``.  Bound for the number of
                         vertices to be tested in the internally constructed simplex.

    - ``cache``  -- A dictionary or ``None``.  For internal usage.
    
    - ``loader`` -- A function which given a level and a weight, returns the
                    a list of sums of extended eta product that span the
                    corresponding space of modular forms.

    OUTPUT:

    - A pair ``(spanned, forms)``, where the first entry indicated whether the whole space
      could be spanned and the second is a list of modular forms.

    EXAMPLES::

        sage: span_with_eta_products(11, 2)
        (True, [1*Lambda_{11,1,0}(z), 1*eta(z)^2 * eta(11z)^2])
    """
    if cache is None :
        cache = dict()

    eta_level = N * helper_level
    
    mf_full = ModularForms(eta_level, k)
    bound = max([ b.valuation() for b in mf_full.echelon_basis() ]) + 1

    print
    print "Spanning level {N} and weight {k} with extended eta products".format(N = N, k = k)
    print "Bound is {0}".format(bound)

    mf = ModularForms(N, k)
    mf_span = span([ vector(QQ, b.qexp(bound).padded_list(bound))
                     for b in mf.basis() ])

    ## Eisenstein series
    if k % 2 == 0 :
        lambda_scales = Subsets( (k // 2) * filter(lambda d: d != 1, Integer(N).divisors()),
                                 k // 2, submultiset = True )
        if k == 2 :
            cur_forms = [ prod(eta_lambda(d) for d in ds) for ds in lambda_scales ]
            cur_span = mf_span.span([ vector(QQ, f.eval_projection(bound).padded_list(bound))
                                      for f in cur_forms ])
        else :
            cur_forms = list()
            cur_span = mf_span.span([])
            for ds in lambda_scales :
                form = prod(eta_lambda(d) for d in ds)
                vec = vector(QQ, form.eval_projection(bound).padded_list(bound))
                if vec not in cur_span :
                    cur_forms.append(form)
                    cur_span += mf_span.span([ vec ])
    else :
        cur_span = mf_span.span([])
        cur_forms = list()

    print "Eta products needed {0}".format(mf.dimension() - len(cur_forms))


    # We span all oldforms, so that we eventually have forms that are
    # as simple as possible
    for dlevel in Integer(N).divisors() :
        if dlevel == N or dlevel == 1 :
            continue

        try :
            nforms = cache[dlevel]
        except KeyError :
            try :
                nforms = loader(dlevel, k)
            except :
                nforms = span_with_eta_products(dlevel, k, helper_level, nmb_vertices, cache, loader)[1]

        for f in nforms :
            for d in Integer(N / dlevel).divisors() :
                rf = f.rescale(d)

                v = vector(QQ, rf.eval_projection(bound).padded_list(bound))
                if v in cur_span :
                    continue

                cur_span = cur_span + span([v])
                cur_forms.append(rf)

    assert cur_span.is_submodule(mf_span)

    if mf.dimension() == len(cur_forms) :
        print "Succesfully spaned level {N} and weight {k} with eta products".format(N = N, k = k)
        print

        cache[N] = cur_forms
        return (True, cur_forms)


    ## Start searching for new forms
    ## Find eta products and sort them in such a way that those which have
    ## fewest terms occur first
    eta_products = sorted( find_eta_products(eta_level, k, True, nmb_vertices),
                           key = lambda e: e.count(0), reverse = True )
    print "Computing possible eta products: Done"

    eta_products = [ eta_product(e, eta_level) for e in eta_products ]
    def check_modularity( e, (a, b, c, d) ) :
        e = e._dict.keys()[0]

        nu = 0
        for (_,M,_,_) in e._e :
            nu += _eta_multiplier( a, M * b, c // M, d )
        if nu not in ZZ :
            return False

        return True

    for g in FareySymbol( Gamma0(eta_level) ).generators() :
        eta_products = filter( lambda e: check_modularity(e, g),
                               eta_products )

    ## Add Eisenstein series of higher level
    if k % 2 == 0 :
        lambda_scales = Subsets( (k // 2) * filter(lambda d: d != 1, Integer(eta_level).divisors()),
                                 k // 2, submultiset = True )
        lambda_scales = filter( lambda ds: any( N % d != 0 for d in ds ),
                                lambda_scales )
        eta_products += [ prod(eta_lambda(d) for d in ds)
                          for ds in lambda_scales ]

    eta_expansions = [ vector(QQ, e._eval_projection__primitive(bound, QQ).padded_list(bound))
                       for e in eta_products ]
    print "Computing preliminary expansions: Done"

    eta_expansion_matrix = matrix(QQ, len(eta_products), bound, eta_expansions).transpose()
    modular_eta_span = mf_span.intersection(eta_expansion_matrix.column_module())
    eta_kernel = eta_expansion_matrix.right_kernel()
    print "Computing intersection and kernel of eta spans: Done"

    for (i, eta_expansion) in enumerate(modular_eta_span.basis()) :
        if eta_expansion in cur_span :
            continue
        cur_span = cur_span + span([eta_expansion])

        eta_coords = eta_expansion_matrix.solve_right(eta_expansion)
        eta_space = eta_kernel + span([eta_coords])
        eta_coords = sorted( eta_space.echelonized_basis(),
                             key = lambda v: v.list().count(0), reverse = True )[0]

        eta = sum( map(operator.mul, eta_coords, eta_products),
                   XEtaProductAlgebra().zero() )
        eta = XEtaProductAlgebra(gcd(N, eta.parent().level()))(eta)
        cur_forms.append( eta )

        if len(cur_forms) == mf.dimension() :
            print "Succesfully spaned level {N} and weight {k} with eta products".format(N = N, k = k)
            print

            cache[N] = cur_forms
            return (True, cur_forms)

    print "UNSuccesfully spaned level {N} and weight {k} with eta products".format(N = N, k = k)
    print

    cache[N] = cur_forms
    return (False, cur_forms)
    
def find_eta_products(N, k, gamma1 = False, nmb_vertices = None) :
    r"""
    Find all holomorphic eta products `\prod_{M \mid N} \eta(M \tau)^{p_M}` of level `N`.

    INPUT:
    
    - `N` -- A positive integer.

    - `k` -- A positive half-integer.
    
    - ``gamma1`` -- A boolean. If ``True`` only eta products with integral exponents
                     at infinity will be returned.

    - ``nmb_vertices`` -- ``None`` or a positive integer. Bound for the number of vertices
                          that are returned.

    OUTPUT:

    - A list of tuples `(p_M)_M` which is sorted by increasing `M`.

    EXAMPLES::

        sage: find_eta_products(6, 1/2)
        [(2, -1, -1, 1), (0, 0, -1, 2), (-1, 2, 0, 0), (1, -1, -1, 2), (0, 1, 0, 0), (1, 0, 0, 0), (2, -1, 0, 0), (0, 0, 0, 1), (-1, 2, 1, -1), (0, 0, 1, 0), (-1, 1, 2, -1), (0, 0, 2, -1)]
    """
    from time import time
    ti = time()

    # The matrix A maps the exponents of the eta products \prod eta(M \tau)^{x_M}
    # to the 24 * N / c.denominator()  * valuation at the cusps
    # r / d, where d | N, and (r, d) = 1 (this is independent of r).
    cusps = [ Cusp(Integer(1) / d) for d in Integer(N).divisors() ]
    A = matrix([ [ N * gcd(M, c.denominator())**2 / Integer(M * c.denominator())
                   for M in Integer(N).divisors() ]
                 for c in cusps])
    Ainv = A.inverse()
    
    
    # We compute the vertices of the simplex that is the intersection of the 'weight hyperplane',
    # \sum x_i = 2 * k, and the preimage of the positive cone under A.
    vertices = [ 2 * k * c / QQ(sum(c)) for c in Ainv.columns() ]
    edges_mat = matrix( [v - vertices[-1] for v in vertices[:-1]] ).transpose()

    # We clear denominator so that we can work over ZZ.
    edges_mat_den = edges_mat.denominator()
    edges_mat = matrix(ZZ, edges_mat * edges_mat_den)
    
    # We reduce the edges to an upper triangular form, and we store the
    # transformation matrix U. This means that the first reduced edge
    # has the form (*, 0, ...) and so on.
    edges = edges_mat.augment(identity_matrix(edges_mat.nrows()))
    edges.echelonize()
    U = edges[:,-edges_mat.nrows():]
    Uinv = U.inverse()
    edges_red = (edges[:,:-edges_mat.nrows()] / edges_mat_den).columns()

    print "prepared: {0}".format(time() - ti)
    ti = time()

    res = [ tuple(Uinv * v) for v in _iterate_eta_products(edges_red, U * vertices[-1], nmb_vertices) ]

    print "iterated: {0}".format(time() - ti)

    if gamma1 :
        return filter( lambda e: sum( d * p for (d, p) in zip(Integer(N).divisors(), e) ) % 24 == 0,
                       res)
    else :
        return res
