from sage.functions.all import ceil


cdef propagate_integrality(v, la, n, v_stack, la_stack, edges_red) :
    r"""
    Take a vector v and change the coordinates v_{n - 1} ... v_{0}
    adding minimal multiples of reduced edges, which are stored in
    ``edges_res``.

    The current values of `v` and `la` are pushed to the corresponding stacks.

    INPUT:

    - `v` -- A tuple of rationals.

    - `la` -- A rational.
    
    - ``v_stack`` -- A list of tuples of rationals.
    
    - ``la_stack`` -- A list of rationals.

    - ``egdes_red`` -- A list of tuples of rationals.
    """
    # first store the current v and la
    v_stack.append(v)
    la_stack.append(la)

    # compute the minimal multiple of the n-th edge that we can
    # add in order to make the n-th coordinate integral.
    la_diff = (ceil(v[n - 1]) - v[n - 1]) / edges_red[n - 1][n - 1]

    v = v + la_diff * edges_red[n - 1]
    la = la + la_diff

    return (v, la)

cpdef _iterate_eta_products(edges_red, initial_v, nmb_vertices) :
    r"""
    Given a set of edges of a simples and a vector, iterate over all integral
    points in this simples.

    INPUT:
    
    - ``edges_red`` -- A list of tuples of rationals, which is assumed to be
                      in upper triangular shape.
    
    - ``initial_v`` -- A tuple of rationals.
    """
    # We construct integral points in the simplex by starting with the last vertex
    # and then adding multiples of edges, \sum \lambda_i e_i, where \sum \lambda_i <= 1.
    # Since edges_red has upper triangular shape, we can determin integral points by
    # considering successively the coordinates.
    # We consider the last coordinate first and the first is considered last.

    # We will make the entries of a vector successively integral.  Then the following
    # multiples of the edges are the only that we may add.
    lambdas_minimal = [1 / e[n] for (n,e) in enumerate(edges_red)]
    edges_minimal = [e * la for (e, la) in zip(edges_red, lambdas_minimal)]

    res = list()
    # Wenn considering the n-th coordinate the stack contains the vectors and
    # corresponding lamdas for the (n + i)-th coordinates.
    v_stack = []
    la_stack = []
    
    # We construct all possible vectors by starting with the n-th vertex of the
    # simples and then successifly add multiples of the edges. Here, we construct
    # an initial vector by adding as small multiples as possible.
    v = initial_v
    la = 0
    n = len(edges_red)
    while n != 0 :
        (v, la) = propagate_integrality(v, la, n, v_stack, la_stack, edges_red)
        if la > 1 :
            v = v_stack.pop()
            la = la_stack.pop()
            break
        else :
            n -= 1
    if n == 0 :
        res.append(v)

    while (n < len(edges_red)) :
        # if we can add the n-the edge without exceeding the bound on lamda
        if la + lambdas_minimal[n] <= 1 :
            v += edges_minimal[n]
            la += lambdas_minimal[n]

            # try to construct a minimal completion of the current vector by
            # adding as small multiples of the edges as possible
            while n > 0:
                (v, la) = propagate_integrality(v, la, n, v_stack, la_stack, edges_red)

                if la <= 1 :
                    n -= 1
                # if the bound is exceeded during the construction of the minimal
                # completion, we go back one index, so that we can increment this
                # coordinate (possibly making it easier to complete the current coordinate
                # to an integer).
                else :
                    v = v_stack.pop()
                    la = la_stack.pop()
                    
                    break
            # if we succeeded in completing the vector, then store it in the list of results.
            if n == 0 :
                res.append(v)
                if nmb_vertices is not None and len(res) >= nmb_vertices :
                    return res

        # if no multiple of n-th edge can be added we step back one index
        else :
            v = v_stack.pop()
            la = la_stack.pop()

            n += 1

    return res
