from sage.all import *
from modular_form_transformations import modular_forms_pivots
from modular_form_transformations__store import ( _stored_inclusion_matrix_level,
                                                  _stored_inclusion_matrix_order,
                                                  _stored_order_isomorphism_matrix,
                                                  _stored_projected_cusp_expansion_matrices,
                                                  load_inclusion_matrix_level,
                                                  load_inclusion_matrix_order,
                                                  load_order_isomorphism_matrix,
                                                  load_projected_cusp_expansion_matrices )

from unittest import skip

def test_inclusion_matrix_level() :
    for (weight, levels) in _stored_inclusion_matrix_level.iteritems() :
        for level1 in levels :
            for level2 in levels :
                if level2 % level1 != 0 or level1 == level2 :
                    continue

                inc_matrix = load_inclusion_matrix_level( level1, level2, weight )
                yield( _test_inclusion_matrix, level1, level2, weight, inc_matrix)

def _test_inclusion_matrix( level1, level2, weight, inc_matrix ) :
    if weight == 0 :
        assert inc_matrix == identity_matrix(QQ, 1)
        return

    mfb1 = ModularForms(level1, weight).echelon_basis()
    mfb2 = ModularForms(level2, weight).echelon_basis()
    prec = 3 * max( b.valuation() for b in mfb2 ) + 20

    for (i, f1) in enumerate(mfb1) :
        f2 = sum(map( operator.mul, inc_matrix.column(i), mfb2 ))
        if f1.qexp(prec) != f2.qexp(prec) :
            raise AssertionError( "Column {0} of inclusion matrix from level {1} to level {2} and weight {3} is defect".format(i, level1, level2, weight) )

def test_inclusion_matrix_order() :
    for ((weight, order1, order2), levels) in _stored_inclusion_matrix_order.iteritems() :
        for level in levels :
            inc_matrix = load_inclusion_matrix_order(level, weight, order1, order2)
            yield (_test_inclusion_matrix_order, level, weight, order1, order2, inc_matrix)

def _test_inclusion_matrix_order(level, weight, order1, order2, inc_matrix) :
    assert order1 <= order2
    order_diff = order2 - order1

    weight1 = weight + 12 * order1
    weight2 = weight + 12 * order2

    if weight1 == 0 :
        pivots1 = [0]
    else :
        pivots1 = modular_forms_pivots(level, weight1)
    pivots2 = modular_forms_pivots(level, weight2)

    if len(pivots1) == 0 :
        assert inc_matrix.ncols() == 0
        return
    prec = 2 * max(max(pivots1), max(pivots2)) + 20 + order_diff

    delta_inv = CuspForms(1, 12).basis()[0].qexp(prec + 1).shift(-1)**-1
    delta_inv_power1 = delta_inv**order1
    delta_inv_power2 = delta_inv**order2

    if weight1 == 0 :
        mfb1 = [1]
    else :
        mfb1 = [ f.qexp(prec) for f in ModularForms(level, weight1).echelon_basis() ]
    mfb2 = [ f.qexp(prec) for f in ModularForms(level, weight2).echelon_basis() ]

    expansion_matrix1 = \
        matrix([ vector(QQ, (delta_inv_power1 * f).padded_list(prec))
                 for f in mfb1 ])
    expansion_matrix1.echelonize()
    if list(expansion_matrix1.pivots()) != pivots1 :
        raise AssertionError( "Pivots for echelon basis of weakly holomorphic modular forms of level {1}, weight {0}, and order {2} are incorrect.".format(weight, level, order1) )
    expansion_matrix2 = \
        matrix([ vector(QQ, (delta_inv_power2 * f).padded_list(prec))
                 for f in mfb2 ])
    expansion_matrix2.echelonize()
    ## To work around a bug in Sage 5.7, Trac #14506???
    expansion_matrix2 = copy(expansion_matrix2)
    if list(expansion_matrix2.pivots()) != pivots2 :
        raise AssertionError( "Pivots for echelon basis of weakly holomorphic modular forms of level {1}, weight {0}, and order {2} are incorrect.".format(weight, level, order2) )


    if expansion_matrix1[:, :-order_diff] != (inc_matrix.transpose() * expansion_matrix2)[:, order_diff:] :
        raise AssertionError( "Incorrect inclusion matrix for weakly holomorphic modular forms of level {0}, weight {1}, and order {2} and {3}.".format(level, weight, order1, order2) )

def test_order_isomorphism_matrix() :
    for ((weight, order), levels) in _stored_order_isomorphism_matrix.iteritems() :
        for level in levels :
            isom_matrix = load_order_isomorphism_matrix(level, weight, order)

            yield (_test_order_isomorphism_matrix, level, weight, order, isom_matrix)

def _test_order_isomorphism_matrix( level, weight, order, isom_matrix ) :
    assert order >= 0

    weight_order = weight + 12 * order
    pivots = modular_forms_pivots(level, weight_order)

    if len(pivots) == 0 :
        prec = 1
    else :
        prec = max(pivots) + 1
    prec = 2 * prec + 20

    delta_power = CuspForms(1, 12).basis()[0].qexp(prec + 1).shift(-1)**order
    delta_inv_power = CuspForms(1, 12).basis()[0].qexp(prec + 1).shift(-1)**(-order)

    mfb = [ f.qexp(prec) for f in ModularForms(level, weight_order).echelon_basis() ]
    
    expansion_matrix = \
        matrix([ vector(QQ, (delta_inv_power * f).padded_list(prec))
                 for f in mfb ])
    expansion_matrix.echelonize()

    R = PowerSeriesRing(QQ, 'q', default_prec = prec)
    mfb_weak = [ R(list(r)) for r in expansion_matrix.rows() ]
    
    isom_matrix_inv = isom_matrix.inverse()

    for (i, f_weak) in enumerate(mfb_weak) :
        f = delta_power * f_weak
        f_img = sum(map( operator.mul, isom_matrix.column(i), mfb ))
        if f[i] != f_img[i] :
            raise AssertionError( "Column {0} of order isomorphism matrix for level {1}, weight {2}, and order {3} is incorrect.".format(i, level, weight, order) )

def test_numerically_stored_projected_cusp_expansion_matrices() :

    for (weight, levels) in _stored_projected_cusp_expansion_matrices.iteritems() :
        for level in levels :
            proj_matrices = load_projected_cusp_expansion_matrices( level, weight )
            yield (_test_numerically_projected_cusp_expansion_matrices, level, weight, proj_matrices)

def _test_numerically_projected_cusp_expansion_matrices(level, weight, proj_matrices, eps = 10**-6) :
    r"""
    By evaluating numerically the associated map, this function tests
    whether the matrix for projected cusp expansions has been computed
    correctly.
    """
    mfb = ModularForms(level, weight).echelon_basis()
    dim = len(mfb)
    for cusp in Gamma0(level).cusps() :
        print cusp
        tmp = proj_matrices(cusp).nrows()
    if any(    dim != proj_matrices(cusp).nrows()
            or dim != proj_matrices(cusp).ncols()
            for cusp in Gamma0(level).cusps() ) :
        raise AssertionError( "Size of matrices for level {0} and weight {1} is incorrect".format(level, weight) )

    for cusp in Gamma0(level).cusps() :
        for i in range(dim) :
            try :
                _test_numerically_projected_cusp_expansion(
                    mfb, i, proj_matrices(cusp).column(i),
                    cusp, eps )
            except AssertionError, err :
                raise AssertionError( "Level {0} and weight {1}: Column {2} for cusp {3} is incorrect.  Numerical evaluation is off by {4}.".format(level, weight, i, cusp, err) )

def _test_numerically_projected_cusp_expansion(mfb, i, g_coords, cusp, eps ) :
    r"""
    Test numerically up to precision ``eps`` whether `\pi_{FE}( f_\frakc )`
    coincides with `g`.
    """
    CC = ComplexField(200)
    (a_g, b_g, c_g, d_g) = _cusp_matrix(cusp)
    gamma = lambda tau: (a_g * tau + b_g) / (c_g * tau + d_g)
    j = lambda tau: CC(c_g * tau + d_g)**-1

    mfb_qexp = [b.qexp(4000).polynomial() for b in mfb]

    cusp_width = mfb[i].parent().group().cusp_data(cusp)[1]

    for tau in [SR('I')] : #, SR('I') + 1/3, SR('2 * I') + 2/5] :
        while(True) :
            f_eval = CC(Integer(1) / cusp_width) * \
                sum(   j(tau + n)**mfb[i].weight()
                       * mfb_qexp[i].subs(q = CC(exp(2 * pi * I * gamma(tau + n))))
                       for n in range(cusp_width) )

            if abs(f_eval) > 10**-4:
                break
            tau = tau / 2

        g_eval = sum([ CC(g_coords[i]) * mfb_qexp[i].subs(q = CC(exp(2 * pi * I * tau)))
                       for i in range(len(mfb)) ])
        
        if abs(f_eval - g_eval) > eps :
            raise AssertionError( str(abs(f_eval - g_eval)) )

def _cusp_matrix( cusp ) :
    if cusp == infinity :
        return (1, 0, 0, 1)
    cups = QQ(cusp)

    a = cusp.numerator()
    c = cusp.denominator()

    (_, b, d) = xgcd(-c, a)

    return (a, b, c, d)
